#define OPTIMIZE_ROM_CALLS
#define NO_AMS_CHECK
#define NO_CALC_DETECT
#define NO_EXIT_SUPPORT
#define USE_TI89
#define USE_TI92PLUS
#define SAVE_SCREEN           
#include <tigcclib.h>


/* Matrix- and vector allocation and freeing is taken from Pate Williams' LLL program
   which can be found at http://www.mindspring.com/~pate/course/chap02/lll.c
*/

float **allocate_real_matrix(int lr, int ur, int lc, int uc)
{
  /* Allocates a real matrix of range [lr..ur][lc..uc]. */
  float **p = calloc((ur - lr + 1), sizeof(float *));
  int i;

  p -= lr;
  for (i = lr; i <= ur; i++) {
    p[i]= calloc((uc - lc + 1), sizeof(float));
	  p[i] -= lc;
  }
  return p;
}


float *allocate_real_vector(int l, int u)
{
  /* Allocates a real vector of range [l..u]. */
  float *p;

  p = calloc((u - l + 1), sizeof(float));
  return p - l;
}

void free_real_matrix(float **m, int lr, int ur, int lc)
{
  /* Frees a real matrix of range [lr..ur][lc..uc]. */
  int i;

  for (i = ur; i >= lr; i--) free((char *)(m[i] + lc));
  free((char *) (m + lr));
}

void free_real_vector(float *v, int l)
{
  /* Frees a real vector of range [l..u]. */

  free((char *)(v + l));
}

float Scalar(long n, float *u, float *v)
{
  /* Calculate the scalar product of two vectors
     [1..n] */
  double sum = 0.0;
  int i;

  for (i = 1; i <= n; i++) sum += u[i] * v[i];
  return sum;
}

void RED(int k, int l, int sjl, float **b, float **mu) {  // Subroutine used in the main LLL function.
	if ( fabs(mu[k][l]) > 0.5 ) {
		int i,j;
		float r;
	  r = floor(0.5 + mu[k][l]);
	  for(i = 1; i <= sjl; i++) {
	  	b[k][i] = b[k][i] - r*b[l][i];
	  }
	  for(j = 1; j <= l-1; j++) {
	  	mu[k][j] = mu[k][j] - r*mu[l][j];
	  }
	  mu[k][l] = mu[k][l] - r;
	  
	
	}
	
}


void LLL(int rk, int sjl, float **b)
{
  /* Lattice reduction algorithm. */
  
  float m,SB,swp,t;
  float *B = allocate_real_vector(1, sjl);
  float **bs = allocate_real_matrix(1, rk, 1, sjl);
  float **mu = allocate_real_matrix(1, rk, 1, sjl);
  int i, j, fv, k, l, k1;
  
  //Initializing...  
  for(i = 1; i <= sjl; i++) {
  	bs[1][i] = b[1][i];
  }
  m = Scalar(sjl, bs[1],bs[1]);
  B[1] = m;
  for(i = 2; i <= rk; i++) {
  	for(fv = 1; fv <= sjl; fv++) {
  		bs[i][fv] = b[i][fv];
  	}
  	for(j=1; j <= i-1; j++) {
  		mu[i][j] = Scalar(sjl,b[i],bs[j])/B[j];
  		for(fv=1; fv <= sjl; fv++) {
  			bs[i][fv] = bs[i][fv] - mu[i][j]*bs[j][fv];
  		}
  	}
  	B[i] = Scalar(sjl,bs[i],bs[i]);
  }
  k = 2;
  //Initialization done!
  
  TrinFire:    //Iteration starts here
  k1 = k-1;
  RED(k,k1,sjl,b,mu);
  if( (B[k] < (0.75 - mu[k][k1]*mu[k][k1])*B[k1]) ) {
  	m = mu[k][k1], SB = B[k] + m*m*B[k1], mu[k][k1] = m*B[k1]/SB, B[k] = B[k1]*B[k]/SB, B[k1] = SB;
  	for(i=1;i<=sjl; i++) {
  		swp = b[k][i];
  		b[k][i] = b[k1][i];
  		b[k1][i] = swp;
  	}
  	if( k > 2 ) {
  		for(j=1;j<=k-2;j++) {
  			swp = mu[k][j];
  			mu[k][j] = mu[k1][j];
  			mu[k1][j] = swp;
  		}
  	}
  	for(i=k+1;i<=rk; i++) {
  		t = mu[i][k], mu[i][k] = mu[i][k1] - m*t, mu[i][k1] = t+mu[k][k1]*mu[i][k];
  	}
  	k = max(2, k1);
  	goto TrinFire;
  }
  for(l=k-2; l>=1; l--) {
  	RED(k,l,sjl,b,mu);
  }
  k = k+1;
  if ( k <= rk ) goto TrinFire;
  
  free_real_vector(B, 1);
  free_real_matrix(bs,1 ,rk ,1 );
  free_real_matrix(mu,1 ,rk ,1 );
  

}

void _main(void)
{
   
   int i,j,AntalVar;
   int NumRow, NumCol;
   AntalVar = ArgCount();
   if ( AntalVar == 0 ) return;   //If no arguments are given, return.
   ESI argptr = top_estack;

   argptr -=2;
   NumCol = remaining_element_count(argptr); //num. of cols.
   NumRow = remaining_element_count(argptr+1); //num. of rows.
   if( NumRow > NumCol || NumRow == 1) return; // Do not accept invalid matrices.
   float **b = allocate_real_matrix(1,NumRow,1,NumCol);   // create matrix
   
   
 
  /* Most of the following code is from Rafael H. Padilla V.'s and Samuel Stearley's program
     Inv.c v1.1  Matrix Inverse Algorith "ANA"
  */
	
	  OSEnableBreak();
	  //take matrix A from the estack.  If there is a symbolic in the estack
    //that can not be approximated an error will be thrown.
    for (i=1; i<=NumRow; i++) {
 	     for (j=1; j<=NumCol; j++) {
				  NG_approxESI(argptr);
				  if (*top_estack != FLOAT_TAG) ER_throw(ER_INVALID_LIST_OR_MATRIX);
   	      b[i][j] = estack_number_to_Float(top_estack);
   	      argptr = next_expression_index(argptr);
				  top_estack = argptr;
   	   }
       argptr -= 2;
    }
    
    LLL(NumRow,NumCol,b);  // Compute LLL-reduced basis for lattice.
 	  
	   // return answer to estack.
	   push_END_TAG();
	   for (i=NumRow; i > 0; i--) {
	     push_END_TAG();
	     for (j=NumCol; j > 0; j--)
	       push_Float(b[i][j]);
	     push_LIST_TAG();
	   }
	   push_LIST_TAG();
		 free_real_matrix(b,1,NumRow,1);		 //free ram before writing file
	   
		 VarStore(SYMSTR("lllred"),STOF_ESI,0,top_estack);
	
		 OSClearBreak();

}